# -*- coding: utf-8 -*-

__author__ = """Ishwar Sawale"""
__email__ = 'ishwar.code@gmail.com'
__version__ = '0.1.0'

from .api import align_input, create_classifier, test_train_classifier, test_classifier
